// RecordWAV.cpp: command-line utility for playing WAV files of unlimited size
//
// This source file is a supplement to the article "Multimedijalne datoteke".
// The article was published in the "PC" computer monthly, published in
// Belgrade, Yugoslavia by "PC Press"
//
// You can freely use, modify and redistribute this file, provided that this
// header is kept unchanged
//
//                                                                 Ivica Nikolic
//                                                      ivica@galeb.etf.bg.ac.yu
//
////////////////////////////////////////////////////////////////////////////////

#include "StdInclude.h"

int nBufferSeconds = 1;
char szFilename[ _MAX_PATH ];
WAVEFORMATEX wfmt;

HMMIO hmmio = NULL;
DWORD dwDataSize = 0;

DWORD dwBufferSize;

void DisplayHelpOnUsage()
{
	puts( "PlayWAV by Ivica Nikolic" );
	puts( "Usage: PlayWAV [options] filename" );
	puts( "" );
	puts( "    options:" );
	puts( "" );
	puts( "        -?      - show this help page" );
	puts( "" );
	puts( "        -bN     - set memory buffer size to N seconds (default 1 second)" );
}

BOOL ParseCommandLine( int argc, char* argv[] )
{
	if ( argc<=1 )
	{
		DisplayHelpOnUsage();
		return FALSE;
	}

	BOOL bFilenamePassed = FALSE;
	BOOL bBufferSizePassed = FALSE;

	BOOL bHelpOnUsageShown = FALSE;

	for ( int nArg=1; nArg<argc; nArg++ )
	{
		const char* szArg = argv[nArg];

		if ( szArg[0] == '-' )
		{
			if ( szArg[1]=='b' )
			{
				if ( bBufferSizePassed )
				{
					puts( "Buffer size attempted to set more than once" );
					return FALSE;
				}

				bBufferSizePassed = TRUE;

				BOOL bOnlyDigits = ( szArg[2] != 0 );
				for ( int i=strlen(szArg)-1; i>=2 && bOnlyDigits; i-- )
					bOnlyDigits &= ( szArg[i]>='0' && szArg[i]<='9' );

				if ( !bOnlyDigits )
				{
					printf( "Invalid command line option: %s\n", szArg );
					return FALSE;
				}

				nBufferSeconds = atol( szArg+2 );
				if ( nBufferSeconds<1 )
				{
					puts( "Buffer can not be shorter than one second" );
					return FALSE;
				}
			}
			else if ( strcmpi(szArg, "-?") == 0 )
			{
				if ( !bHelpOnUsageShown )
				{
					DisplayHelpOnUsage();
					bHelpOnUsageShown = TRUE;
				}
			}
			else
			{
				printf( "Invalid command line option: %s\n", szArg );
				return FALSE;
			}
		}
		else
		{
			if ( bFilenamePassed )
			{
				puts( "Only one filename is allowed" );
				return FALSE;
			}

			if ( strlen(szArg) > _MAX_PATH-1 )
			{
				puts( "Filename too long" );
				return FALSE;
			}

			strcpy( szFilename, szArg );
			bFilenamePassed = TRUE;
		}
	}

	if ( !bFilenamePassed && !bHelpOnUsageShown )
	{
		puts( "No filename given" );
		return FALSE;
	}

	return TRUE;
}

BOOL OpenFileAndLocateDataChunk()
{
	MMRESULT mmres;
	DWORD dwBytesRead;

	MMCKINFO mmckinfoRIFF;
	MMCKINFO mmckinfoFMT;
	MMCKINFO mmckinfoDATA;

	DWORD dwFmtSize;

	hmmio = mmioOpen( szFilename, NULL, MMIO_READ | MMIO_ALLOCBUF );
	if ( hmmio==NULL )
	{
		puts( "Can not open file" );
		goto labelSomethingWentWrong;
	}

	mmckinfoRIFF.fccType = mmioFOURCC( 'W', 'A', 'V', 'E' );
	mmres = mmioDescend( hmmio, &mmckinfoRIFF, NULL, MMIO_FINDRIFF );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "File does not contain waveform audio" );
		goto labelSomethingWentWrong;
	}

	mmckinfoFMT.ckid = mmioFOURCC( 'f', 'm', 't', ' ' );
	mmres = mmioDescend( hmmio, &mmckinfoFMT, &mmckinfoRIFF,
		MMIO_FINDCHUNK );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Mangled file - can not locate format information" );
		goto labelSomethingWentWrong;
	}

	dwFmtSize = mmckinfoFMT.cksize;
	if ( dwFmtSize < sizeof(wfmt)-sizeof(wfmt.cbSize) )
	{
		puts( "Mangled file - format information is truncated" );
		goto labelSomethingWentWrong;
	}

	if ( dwFmtSize > sizeof(wfmt) )
		dwFmtSize = sizeof(wfmt);

	dwBytesRead = mmioRead( hmmio, (HPSTR)&wfmt, dwFmtSize );
	if ( dwBytesRead != dwFmtSize )
	{
		puts( "An error occured while reading the file" );
		goto labelSomethingWentWrong;
	}

	if ( wfmt.wFormatTag != WAVE_FORMAT_PCM )
	{
		puts( "Audio data format is not WAVE_FORMAT_PCM" );
		goto labelSomethingWentWrong;
	}

	if ( dwFmtSize < sizeof(WAVEFORMAT) )
		wfmt.wBitsPerSample = 8 * wfmt.nBlockAlign / wfmt.nChannels;

	wfmt.cbSize = 0;

	HWAVEOUT hwoDummy;
	mmres = waveOutOpen( &hwoDummy, WAVE_MAPPER, &wfmt, 0, 0L,
		WAVE_FORMAT_QUERY );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Audio data format not supported by hardware" );
		goto labelSomethingWentWrong;
	}

	mmres = mmioAscend( hmmio, &mmckinfoFMT, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "An error occured while reading the file" );
		goto labelSomethingWentWrong;
	}

	mmckinfoDATA.ckid = mmioFOURCC( 'd', 'a', 't', 'a' );
	mmres = mmioDescend( hmmio, &mmckinfoDATA, &mmckinfoRIFF, MMIO_FINDCHUNK );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Mangled file - can not locate waveform data" );
		goto labelSomethingWentWrong;
	}

	dwDataSize = mmckinfoDATA.cksize;

	dwBufferSize = nBufferSeconds * wfmt.nSamplesPerSec * wfmt.nBlockAlign;

	return TRUE;

labelSomethingWentWrong:
	if ( hmmio!=NULL )
		mmioClose( hmmio, NULL );

	return FALSE;
}

BOOL ReadAndSendBuffer( HWAVEOUT hwo, WAVEHDR* pwhdr, LONG nBytes )
{
	pwhdr->dwFlags = 0;

	LONG nBytesRead =  mmioRead( hmmio, pwhdr->lpData, nBytes );
	if ( nBytesRead != nBytes )
	{
		puts( "Error reading file" );
		return FALSE;
	}

	pwhdr->dwBufferLength = (DWORD)nBytes;

	waveOutPrepareHeader( hwo, pwhdr, sizeof(*pwhdr) );

	MMRESULT mmres = waveOutWrite( hwo, pwhdr, sizeof(*pwhdr) );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error writing to wave output device" );
		return FALSE;
	}

	return TRUE;
}

void PlaySample()
{
	MMRESULT mmres;

	HWAVEOUT hwo = NULL;

	WAVEHDR whdr1;
	WAVEHDR whdr2;

	DWORD dwRemainingBytes;
	DWORD dwBytesToRead;

	BOOL bShouldStopPlayback = FALSE;

	memset( &whdr1, 0, sizeof(whdr1) );
	memset( &whdr2, 0, sizeof(whdr2) );

	whdr1.lpData = (LPSTR)GlobalAllocPtr( GHND, dwBufferSize );
	whdr2.lpData = (LPSTR)GlobalAllocPtr( GHND, dwBufferSize );

	if ( whdr1.lpData==NULL || whdr2.lpData==NULL )
		goto labelFinish;

	mmres = waveOutOpen( &hwo, WAVE_MAPPER, &wfmt, 0, 0, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error - can not open wave out device" );
		goto labelFinish;
	}

	dwRemainingBytes = dwDataSize;
	dwBytesToRead =
		dwRemainingBytes>=dwBufferSize ? dwBufferSize : dwRemainingBytes;

	if ( !ReadAndSendBuffer(hwo, &whdr1, dwBytesToRead) )
		goto labelFinish;

	dwRemainingBytes -= dwBytesToRead;

	whdr2.dwFlags |= WHDR_DONE;

	printf( "Playback started: sampling rate %dHz, %d bits per sample, %d channel(s)\n",
		wfmt.nSamplesPerSec, wfmt.wBitsPerSample, wfmt.nChannels );
	puts( "Press space to stop" );

	while ( dwRemainingBytes != 0 && !bShouldStopPlayback )
	{
		WAVEHDR* pwhdr = NULL;

		if ( whdr1.dwFlags & WHDR_DONE )
			pwhdr = &whdr1;
		else if ( whdr2.dwFlags & WHDR_DONE )
			pwhdr = &whdr2;

		if ( pwhdr!=NULL )
		{
			dwBytesToRead =
				dwRemainingBytes>=dwBufferSize ? dwBufferSize : dwRemainingBytes;

			if ( pwhdr->dwFlags & WHDR_PREPARED )
				waveOutUnprepareHeader( hwo, pwhdr, sizeof(*pwhdr) );
			if ( !ReadAndSendBuffer(hwo, pwhdr, dwBytesToRead) )
				goto labelFinish;

			dwRemainingBytes -= dwBytesToRead;
		}

		if ( _kbhit() )
		{
			if ( _getch()==' ' )
			{
				bShouldStopPlayback = TRUE;
				continue;
			}
		}

		Sleep( 50 );
	}

	if ( !bShouldStopPlayback )
	{
		while ( !(whdr1.dwFlags & WHDR_DONE) || !(whdr2.dwFlags & WHDR_DONE) )
			Sleep( 50 );
	}

labelFinish:

	if ( hwo != NULL )
	{
		waveOutReset( hwo );

		if ( whdr1.dwFlags & WHDR_PREPARED )
			waveOutUnprepareHeader( hwo, &whdr1, sizeof(whdr1) );
		if ( whdr2.dwFlags & WHDR_PREPARED )
			waveOutUnprepareHeader( hwo, &whdr2, sizeof(whdr2) );

		waveOutClose( hwo );
	}

	if ( whdr1.lpData != NULL )
		GlobalFreePtr( whdr1.lpData );
	if ( whdr2.lpData != NULL )
		GlobalFreePtr( whdr2.lpData );
}

void main( int argc, char* argv[] )
{
	if ( !ParseCommandLine(argc, argv) )
		return;

	if ( !OpenFileAndLocateDataChunk() )
		return;

	PlaySample();

	mmioClose( hmmio, NULL );
}
