// RecordWAV.cpp: command-line utility for recording WAV files of unlimited size
//
// This source file is a supplement to the article "Multimedijalne datoteke".
// The article was published in the "PC" computer monthly, published in
// Belgrade, Yugoslavia by "PC Press"
//
// You can freely use, modify and redistribute this file, provided that this
// header is kept unchanged
//
//                                                                 Ivica Nikolic
//                                                      ivica@galeb.etf.bg.ac.yu
//
////////////////////////////////////////////////////////////////////////////////

#include "StdInclude.h"

DWORD dwFSamp = 8000;
WORD wBitsPerSample = 8;
WORD nChannels = 1;
int nBufferSeconds = 1;
char szFilename[ _MAX_PATH ];
WAVEFORMATEX wfmt;

HMMIO hmmio = NULL;

DWORD dwBufferSize;

MMCKINFO mmckinfoRIFF;
MMCKINFO mmckinfoDATA;

void DisplayHelpOnUsage()
{
	puts( "RecordWAV by Ivica Nikolic" );
	puts( "Usage: RecordWAV [options] filename" );
	puts( "" );
	puts( "    options:" );
	puts( "" );
	puts( "        -?      - show this help page" );
	puts( "" );
	puts( "        -8000   - use  8000Hz sampling rate           (default)" );
	puts( "        -11025  - use 11025Hz sampling rate" );
	puts( "        -22050  - use 22050Hz sampling rate" );
	puts( "        -44100  - use 44100Hz sampling rate" );
	puts( "" );
	puts( "        -8      -  8 bits per sample                  (default)" );
	puts( "        -16     - 16 bits per sample" );
	puts( "" );
	puts( "        -mono   - record in mono mode                 (default)" );
	puts( "        -stereo - record in stereo mode" );
	puts( "" );
	puts( "        -bN     - set memory buffer size to N seconds (default 1 second)" );
}

BOOL ParseCommandLine( int argc, char* argv[] )
{
	if ( argc<=1 )
	{
		DisplayHelpOnUsage();
		return FALSE;
	}

	BOOL bFSampPassed = FALSE;
	BOOL bBitsPerSamplePassed = FALSE;
	BOOL bNChannelsPassed = FALSE;
	BOOL bFilenamePassed = FALSE;
	BOOL bBufferSizePassed = FALSE;

	BOOL bHelpOnUsageShown = FALSE;

	for ( int nArg=1; nArg<argc; nArg++ )
	{
		const char* szArg = argv[nArg];

		if ( szArg[0] == '-' )
		{
			if ( szArg[1]=='b' )
			{
				if ( bBufferSizePassed )
				{
					puts( "Buffer size attempted to set more than once" );
					return FALSE;
				}

				bBufferSizePassed = TRUE;

				BOOL bOnlyDigits = ( szArg[2] != 0 );
				for ( int i=strlen(szArg)-1; i>=2 && bOnlyDigits; i-- )
					bOnlyDigits &= ( szArg[i]>='0' && szArg[i]<='9' );

				if ( !bOnlyDigits )
				{
					printf( "Invalid command line option: %s\n", szArg );
					return FALSE;
				}

				nBufferSeconds = atol( szArg+2 );
				if ( nBufferSeconds<1 )
				{
					puts( "Buffer can not be shorter than one second" );
					return FALSE;
				}
			}
			else if (
				strcmpi(szArg, "-8000") == 0 ||
				strcmpi(szArg, "-11025") == 0 ||
				strcmpi(szArg, "-22050") == 0 ||
				strcmpi(szArg, "-44100") == 0
			) {
				if ( bFSampPassed )
				{
					puts( "Sampling rate attempted to set more than once" );
					return FALSE;
				}

				dwFSamp = atol( szArg+1 );

				bFSampPassed = TRUE;
			}
			else if (
				strcmpi(szArg, "-8") == 0 ||
				strcmpi(szArg, "-16") == 0
			) {
				if ( bBitsPerSamplePassed )
				{
					puts( "Bits per sample attempted to set more than once" );
					return FALSE;
				}

				wBitsPerSample = (WORD)atol( szArg+1 );

				bBitsPerSamplePassed = TRUE;
			}
			else if (
				strcmpi(szArg, "-mono") == 0 ||
				strcmpi(szArg, "-stereo") == 0
			) {
				if ( bNChannelsPassed )
				{
					puts( "Number of channels attempted to set more than once" );
					return FALSE;
				}

				if ( strcmpi(szArg, "-mono") == 0 )
					nChannels = 1;
				else
					nChannels = 2;

				bNChannelsPassed = TRUE;
			}
			else if ( strcmpi(szArg, "-?") == 0 )
			{
				if ( !bHelpOnUsageShown )
				{
					DisplayHelpOnUsage();
					bHelpOnUsageShown = TRUE;
				}
			}
			else
			{
				printf( "Invalid command line option: %s\n", szArg );
				return FALSE;
			}
		}
		else
		{
			if ( bFilenamePassed )
			{
				puts( "Only one filename is allowed" );
				return FALSE;
			}

			if ( strlen(szArg) > _MAX_PATH-1 )
			{
				puts( "Filename too long" );
				return FALSE;
			}

			strcpy( szFilename, szArg );
			bFilenamePassed = TRUE;
		}
	}

	if ( !bFilenamePassed && !bHelpOnUsageShown )
	{
		puts( "No filename given" );
		return FALSE;
	}


	return TRUE;
}

BOOL CreateFileAndChunks()
{
	MMRESULT mmres;
	LONG nBytesWritten;

	MMCKINFO mmckinfoFMT;

	wfmt.wFormatTag = WAVE_FORMAT_PCM;
	wfmt.nChannels = nChannels;
	wfmt.nSamplesPerSec = dwFSamp;
	wfmt.nAvgBytesPerSec = dwFSamp * nChannels * wBitsPerSample / 8;
	wfmt.nBlockAlign = nChannels * wBitsPerSample / 8;
	wfmt.wBitsPerSample = wBitsPerSample;
	wfmt.cbSize = 0;

	HWAVEIN hwiDummy;
	mmres = waveInOpen( &hwiDummy, WAVE_MAPPER, &wfmt, 0, 0L,
		WAVE_FORMAT_QUERY );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Audio data format not supported by hardware" );
		goto labelSomethingWentWrong;
	}


	dwBufferSize = nBufferSeconds * wfmt.nAvgBytesPerSec;

	hmmio = mmioOpen( szFilename, NULL,
		MMIO_ALLOCBUF | MMIO_WRITE | MMIO_CREATE );
	if ( hmmio == NULL )
	{
		puts( "Can not create file" );
		goto labelSomethingWentWrong;
	}

	mmckinfoRIFF.fccType = mmioFOURCC('W', 'A', 'V', 'E');
	mmres = mmioCreateChunk( hmmio, &mmckinfoRIFF, MMIO_CREATERIFF );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Can not create RIFF chunk" );
		goto labelSomethingWentWrong;
	}

	mmckinfoFMT.ckid = mmioFOURCC('f', 'm', 't', ' ');
	mmckinfoFMT.cksize = sizeof(wfmt);
	mmres = mmioCreateChunk( hmmio, &mmckinfoFMT, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Can not create FMT chunk" );
		goto labelSomethingWentWrong;
	}

	nBytesWritten = mmioWrite( hmmio, (HPSTR)&wfmt, sizeof(wfmt) );
	if ( nBytesWritten != sizeof(wfmt) )
	{
		puts( "Error writing file" );
		goto labelSomethingWentWrong;
	}

	mmres = mmioAscend( hmmio, &mmckinfoFMT, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error writing file" );
		goto labelSomethingWentWrong;
	}

	mmckinfoDATA.ckid = mmioFOURCC('d', 'a', 't', 'a');
	mmres = mmioCreateChunk( hmmio, &mmckinfoDATA, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Can not create DATA chunk" );
		goto labelSomethingWentWrong;
	}

	return TRUE;

labelSomethingWentWrong:
	if ( hmmio!=NULL )
		mmioClose( hmmio, NULL );

	return FALSE;
}

BOOL PrepareAndAddBuffer( HWAVEIN hwi, WAVEHDR* pwhdr )
{
	pwhdr->dwFlags = 0;
	pwhdr->dwBufferLength = dwBufferSize;

	waveInPrepareHeader( hwi, pwhdr, sizeof(*pwhdr) );

	MMRESULT mmres = waveInAddBuffer( hwi, pwhdr, sizeof(*pwhdr) );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error writing to wave output device" );
		return FALSE;
	}

	return TRUE;
}

BOOL UnprepareAndWriteBuffer( HWAVEIN hwi, WAVEHDR* pwhdr )
{
	waveInUnprepareHeader( hwi, pwhdr, sizeof(pwhdr) );

	LONG nBytesWritten =  mmioWrite( hmmio, pwhdr->lpData,
		pwhdr->dwBytesRecorded );
	if ( nBytesWritten != (LONG)pwhdr->dwBytesRecorded )
	{
		puts( "Error writing file" );
		return FALSE;
	}

	pwhdr->dwFlags = 0;

	return TRUE;
}

void RecordSample()
{
	MMRESULT mmres;

	HWAVEIN hwi = NULL;

	WAVEHDR whdr1;
	WAVEHDR whdr2;
	WAVEHDR* pwhdr = NULL;

	BOOL bShouldStopRecording = FALSE;

	memset( &whdr1, 0, sizeof(whdr1) );
	memset( &whdr2, 0, sizeof(whdr2) );

	whdr1.lpData = (LPSTR)GlobalAllocPtr( GHND, dwBufferSize );
	whdr2.lpData = (LPSTR)GlobalAllocPtr( GHND, dwBufferSize );

	if ( whdr1.lpData==NULL || whdr2.lpData==NULL )
		goto labelFinish;

	mmres = waveInOpen( &hwi, WAVE_MAPPER, &wfmt, 0, 0, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error - can not open wave in device" );
		goto labelFinish;
	}

	if ( !PrepareAndAddBuffer(hwi, &whdr1) )
		goto labelFinish;

	whdr2.dwFlags |= WHDR_DONE;

	puts( "Press space to start recording" );
	while ( _getch() != ' ' )
		;

	mmres = waveInStart( hwi );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error - can not start recording" );
		goto labelFinish;
	}

	printf( "Recording started: sampling rate %dHz, %d bits per sample, %d channel(s)\n",
		wfmt.nSamplesPerSec, wfmt.wBitsPerSample, wfmt.nChannels );
	puts( "Press space to stop" );

	while ( !bShouldStopRecording )
	{
		pwhdr = NULL;

		if ( whdr1.dwFlags & WHDR_DONE )
			pwhdr = &whdr1;
		else if ( whdr2.dwFlags & WHDR_DONE )
			pwhdr = &whdr2;

		if ( pwhdr!=NULL )
		{
			if ( pwhdr->dwFlags & WHDR_PREPARED )
				if ( !UnprepareAndWriteBuffer(hwi, pwhdr) )
					goto labelFinish;

			if ( !PrepareAndAddBuffer(hwi, pwhdr) )
				goto labelFinish;
		}

		if ( _kbhit() )
		{
			if ( _getch()==' ' )
			{
				bShouldStopRecording = TRUE;
				continue;
			}
		}

		Sleep( 50 );
	}

	waveInStop( hwi );

	pwhdr = NULL;

	if ( (whdr1.dwFlags & WHDR_DONE) && (whdr1.dwFlags & WHDR_PREPARED) )
		pwhdr = &whdr1;
	else if ( (whdr2.dwFlags & WHDR_DONE) && (whdr2.dwFlags & WHDR_PREPARED) )
		pwhdr = &whdr2;

	if ( pwhdr!=NULL )
		UnprepareAndWriteBuffer( hwi, pwhdr );

labelFinish:

	if ( hwi != NULL )
	{
		waveInReset( hwi );

		if ( whdr1.dwFlags & WHDR_PREPARED )
			waveInUnprepareHeader( hwi, &whdr1, sizeof(whdr1) );
		if ( whdr2.dwFlags & WHDR_PREPARED )
			waveInUnprepareHeader( hwi, &whdr2, sizeof(whdr2) );

		waveInClose( hwi );
	}

	mmres = mmioAscend( hmmio, &mmckinfoDATA, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error writing file" );
	}

	mmres = mmioAscend( hmmio, &mmckinfoRIFF, 0 );
	if ( mmres != MMSYSERR_NOERROR )
	{
		puts( "Error writing file" );
	}
}

void main( int argc, char* argv[] )
{
	if ( !ParseCommandLine(argc, argv) )
		return;

	if ( !CreateFileAndChunks() )
		return;

	RecordSample();

	mmioClose( hmmio, NULL );
}
